/*
           Copyright (C) 2012 Gilles Landais (CDS)

    Author: Gilles Landais, Strasbourg astronomical Data Center (CDS)
    Email: gilles.landais@unistra.fr

    This file is part of H3C.

    H3C is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    H3C is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with H3C; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

/** @file h3c.c
 *  @brief routine healpix for PostgreSQL
 *  @author G.Landais (CDS)
 *  @date 19/jul/2011
 *
 *  @mainpage About H3C
 *
 *  H3C is a PostgreSQL library dedicated for astronomy.
 *  The Library contains functions to index positions (ra,de) with the HEALPix
 *  and others to make a cone search or a crossmatch, it can work with circle, 
 *  polygon or box.
 *
 *  - we use the healpix library provided by the Nasa
 *      (see: http://healpix.jpl.nasa.gov/index.shtml)
 *  - the library is inpired from Q3C 
 *          (see http://code.google.com/p/q3c/)
 *  - the library use the NESTED numerotation  (see NSIDE)
 *  
 *  @see @ref install 
 *  @see @ref licence GPL licence
 *  @see @ref notes
 *  @see @ref howto
 */
/** @page install Installation guide
 *  @include README.h3c
 */
/** @page licence Licence
 *  @include COPYING
 */
/** @page notes H3C tuning
 *  @include Notes.txt
 */
/** @page howto SQL functions
 *  @include HOWTO
 */


#include <string.h>

/* Postgres stuff */
#include "postgres.h"
#include "executor/spi.h"
#include "funcapi.h"
#include "utils/lsyscache.h"
#include "utils/array.h"
#include "catalog/pg_type.h"

#if PG_VERSION_NUM >= 80300
#include "access/tupmacs.h"
#endif

/* End of Postgres stuff */

#include "common.h"
#include "h3c_util.h"

#define LOG_NPASS 200000

/* For PostgreSQL versions >= 8.2 */
#ifdef PG_MODULE_MAGIC
PG_MODULE_MAGIC;
#endif


/* Postgres functions */
Datum pgh3c_version(PG_FUNCTION_ARGS);
Datum pgh3c_ang2ipix(PG_FUNCTION_ARGS);
Datum pgh3c_ang2ipix_default(PG_FUNCTION_ARGS);
Datum pgh3c_ipix2ang(PG_FUNCTION_ARGS);
Datum pgh3c_ipix2ang_default(PG_FUNCTION_ARGS);
Datum pgh3c_disk_ipix(PG_FUNCTION_ARGS);
Datum pgh3c_disk_ipix_default(PG_FUNCTION_ARGS);
Datum pgh3c_radial_query_it(PG_FUNCTION_ARGS);
Datum pgh3c_radial_query_it_default(PG_FUNCTION_ARGS);
Datum pgh3c_dist(PG_FUNCTION_ARGS);
Datum pgh3c_sindist(PG_FUNCTION_ARGS);
Datum pgh3c_poly_query_it(PG_FUNCTION_ARGS);
Datum pgh3c_poly_query_it_default(PG_FUNCTION_ARGS);
Datum pgh3c_in_poly_convex(PG_FUNCTION_ARGS);
Datum pgh3c_in_poly(PG_FUNCTION_ARGS);
/*Datum pgh3c_in_box(PG_FUNCTION_ARGS);*/
Datum pgh3c_poly_center(PG_FUNCTION_ARGS);
Datum pgh3c_circle_intersect_poly(PG_FUNCTION_ARGS);
Datum pgh3c_poly_intersect_poly(PG_FUNCTION_ARGS);
/*Datum pgh3c_box_query_it(PG_FUNCTION_ARGS);
Datum pgh3c_box_query_it_default(PG_FUNCTION_ARGS);*/
Datum pgh3c_box_to_poly(PG_FUNCTION_ARGS);

Datum pgh3c_nearby_it(PG_FUNCTION_ARGS);
Datum pgh3c_nearby_it_default(PG_FUNCTION_ARGS);
Datum pgh3c_dist_ipix(PG_FUNCTION_ARGS);
Datum pgh3c_in_ellipse(PG_FUNCTION_ARGS);
Datum pgh3c_ellipse_query_it(PG_FUNCTION_ARGS);

Datum pgh3c_circle_area(PG_FUNCTION_ARGS);
Datum pgh3c_poly_area(PG_FUNCTION_ARGS);
Datum pgh3c_ipix2default_it(PG_FUNCTION_ARGS);

/*****************************************************************************/
/** @brief the version 
*/
/*****************************************************************************/
PG_FUNCTION_INFO_V1(pgh3c_version);
Datum pgh3c_version(PG_FUNCTION_ARGS)
{
#ifdef VERSION
    PG_RETURN_FLOAT8(VERSION);
#else
    PG_RETURN_NULL();
#endif
}

/*****************************************************************************/
/** @brief (PostgreSQL function) calculate ipix from coordinate
 *  @param  ra right ascension
 *  @param  dec declination
 *  @param  nside nside level
 *  @return ipix
 */
/*****************************************************************************/
PG_FUNCTION_INFO_V1(pgh3c_ang2ipix);
Datum pgh3c_ang2ipix(PG_FUNCTION_ARGS)
{
        h3c_coord_t ra = PG_GETARG_FLOAT8(0);
        h3c_coord_t dec = PG_GETARG_FLOAT8(1);
        int nside = PG_GETARG_UINT32(2);
        h3c_ipix_t ipix;
        static int invocation;
        static h3c_coord_t ra_buf, dec_buf;
        static h3c_ipix_t ipix_buf;
        static int nside_buf;

        if (invocation == 0)
        {

        }
        else
        {
                if ((ra == ra_buf) && (dec == dec_buf) && nside == nside_buf)
                {
                        PG_RETURN_INT64(ipix_buf);
                }
        }

        h3c_ang2pix_nest(nside,
                     h3c_dec2theta(dec*H3C_DEGRA),
                     h3c_ra2phi(ra*H3C_DEGRA),
                     (h3c_ipix_t *) &ipix);

        ra_buf = ra;
        dec_buf = dec;
        ipix_buf = ipix;
        nside_buf=nside;
        invocation=1;
        PG_RETURN_INT64(ipix);
}

/*****************************************************************************/
/** @brief (PostgreSQL function) calculate ipix from coordinate
 *  @param  ra right ascension
 *  @param  dec declination
 *  @return ipix
 *
 *  @note default nside=H3C_DEFAULT_NSIDE
 */
/*****************************************************************************/
PG_FUNCTION_INFO_V1(pgh3c_ang2ipix_default);
Datum pgh3c_ang2ipix_default(PG_FUNCTION_ARGS)
{
        h3c_coord_t ra = PG_GETARG_FLOAT8(0);
        h3c_coord_t dec = PG_GETARG_FLOAT8(1);
        h3c_ipix_t ipix;
        static int invocation;
        static h3c_coord_t ra_buf, dec_buf;
        static h3c_ipix_t ipix_buf;

        if (invocation == 0)
        {

        }
        else
        {
                if ((ra == ra_buf) && (dec == dec_buf))
                {
                        PG_RETURN_INT64(ipix_buf);
                }
        }

        h3c_ang2pix_nest(H3C_DEFAULT_NSIDE,
                     h3c_dec2theta(dec*H3C_DEGRA),
                     h3c_ra2phi(ra*H3C_DEGRA),
                     (h3c_ipix_t *) &ipix);

        ra_buf = ra;
        dec_buf = dec;
        ipix_buf = ipix;
        invocation=1;
        PG_RETURN_INT64(ipix);
}

/*****************************************************************************/
/** @brief (PostgreSQL function) get the center of an ipix
 *  @param  ipix healpix number
 *  @param  nside nside level
 *  @return array (ra,dec)
 */
/*****************************************************************************/
PG_FUNCTION_INFO_V1(pgh3c_ipix2ang);
Datum pgh3c_ipix2ang(PG_FUNCTION_ARGS)
{
        h3c_ipix_t ipix;
        h3c_coord_t theta, phi;

        Datum       *data;
        int16       typlen;
        bool        typbyval;
        char        typalign;
        ArrayType  *result;
        ipix = PG_GETARG_INT64(0);
        h3c_pix2ang_nest(PG_GETARG_INT32(1),
                     ipix,
                     &theta,
                     &phi);
        
       theta=h3c_theta2dec(theta)*H3C_RADEG;
       phi=h3c_phi2ra(phi)*H3C_RADEG;

        /* we have one dimension */
        data = ( Datum *) palloc(sizeof(Datum)*2);
        data[0] = Float8GetDatum (phi);
        data[1] = Float8GetDatum (theta);
        /* get required info about the element type */
        get_typlenbyvalalign(FLOAT8OID, &typlen, &typbyval, &typalign);

        /* now build the array */
        result = construct_array(data, 2, FLOAT8OID, typlen, typbyval, typalign);

        PG_RETURN_ARRAYTYPE_P(result);
}

/*****************************************************************************/
/** @brief (PostgreSQL function) get the center of an ipix
 *  @param  ipix healpix number
 *  @return array (ra,dec)
 *
 *  @note default nside=H3C_DEFAULT_NSIDE
 */
/*****************************************************************************/
PG_FUNCTION_INFO_V1(pgh3c_ipix2ang_default);
Datum pgh3c_ipix2ang_default(PG_FUNCTION_ARGS)
{
        h3c_ipix_t ipix;
        h3c_coord_t theta, phi;

        Datum       *data;
        int16       typlen;
        bool        typbyval;
        char        typalign;
        ArrayType  *result;
        ipix = PG_GETARG_INT64(0);
        h3c_pix2ang_nest(H3C_DEFAULT_NSIDE,
                     ipix,
                     &theta,
                     &phi);
       theta=h3c_theta2dec(theta)*H3C_RADEG;
       phi=h3c_phi2ra(phi)*H3C_RADEG;

        /* we have one dimension */
        data = ( Datum *) palloc(sizeof(Datum)*2);
        data[0] = Float8GetDatum (phi);
        data[1] = Float8GetDatum (theta);
        /* get required info about the element type */
        get_typlenbyvalalign(FLOAT8OID, &typlen, &typbyval, &typalign);

        /* now build the array */
        result = construct_array(data, 2, FLOAT8OID, typlen, typbyval, typalign);

        PG_RETURN_ARRAYTYPE_P(result);
}

/*****************************************************************************/
/** @brief (PostgreSQL function) get the ipix list of a disk
 *  @param ra right ascension
 *  @param dec declination
 *  @param radius radius of the disk
 *  @param nside nside level
 *  @return a set of ipix
 */
/*****************************************************************************/
PG_FUNCTION_INFO_V1(pgh3c_disk_ipix);
Datum pgh3c_disk_ipix(PG_FUNCTION_ARGS)
{
  FuncCallContext *funcctx;
  int call_cntr, count;
  static h3c_ipix_t *ipix;

        if (SRF_IS_FIRSTCALL()) {
                MemoryContext oldcontext;
            h3c_debug_init_clock();

            /* create a function context for cross-call persistence
             */
            funcctx = SRF_FIRSTCALL_INIT();

            /* switch to memory context appropriate for multiple
               function calls */
            oldcontext =
                MemoryContextSwitchTo(funcctx->
                              multi_call_memory_ctx);

            MemoryContextSwitchTo(oldcontext);

                ipix = h3c_disk_ipix(PG_GETARG_FLOAT8(0),
                                 PG_GETARG_FLOAT8(1),
                                 PG_GETARG_FLOAT8(2),
                                 PG_GETARG_UINT32(3),
                                 &count) ;
        }

        /* stuff done on every call of the function */
        funcctx = SRF_PERCALL_SETUP();

        call_cntr = funcctx->call_cntr;

        h3c_log(H3C_DEBUG, "pgh3c_disk_ipix end");
        if (ipix && ipix[call_cntr] > -1) {  /* do when there is more left to send */
            if (ipix[call_cntr] < 0) ipix[call_cntr] = ipix[call_cntr]*-1;
            SRF_RETURN_NEXT(funcctx,Int64GetDatum(ipix[call_cntr]));
        } 
        else {  /* do when there is no more left */
            free(ipix);
            SRF_RETURN_DONE(funcctx);
        }
}

/*****************************************************************************/
/** @brief (PostgreSQL function) get the ipix list of a disk
 *  @param ra right ascension
 *  @param dec declination
 *  @param radius radius of the disk
 *  @return a set of ipix
 *
 *  @note
 *  - default nside=H3C_DEFAULT_NSIDE
 *  - list all the ipix
 */
/*****************************************************************************/
PG_FUNCTION_INFO_V1(pgh3c_disk_ipix_default);
Datum pgh3c_disk_ipix_default(PG_FUNCTION_ARGS)
{  
  FuncCallContext *funcctx;
  int call_cntr, count;
  static h3c_ipix_t *ipix;

        if (SRF_IS_FIRSTCALL()) {
                MemoryContext oldcontext;
            h3c_debug_init_clock();

            /* create a function context for cross-call persistence
             */
            funcctx = SRF_FIRSTCALL_INIT();

            /* switch to memory context appropriate for multiple
               function calls */
            oldcontext =
                MemoryContextSwitchTo(funcctx->
                              multi_call_memory_ctx);

            MemoryContextSwitchTo(oldcontext);

                ipix = h3c_disk_ipix(PG_GETARG_FLOAT8(0),
                                 PG_GETARG_FLOAT8(1),
                                 PG_GETARG_FLOAT8(2),
                                 H3C_DEFAULT_NSIDE,
                                 &count) ;
        }

        /* stuff done on every call of the function */
        funcctx = SRF_PERCALL_SETUP();

        call_cntr = funcctx->call_cntr;

        h3c_log(H3C_DEBUG, "pgh3c_disk_ipix_default end");
        if (ipix && ipix[call_cntr] > -1) {  /* do when there is more left to send */
            SRF_RETURN_NEXT(funcctx,Int64GetDatum(ipix[call_cntr]));
        }
        else {  /* do when there is no more left */
            free(ipix);
            SRF_RETURN_DONE(funcctx);
        }
}

/*****************************************************************************/
/** @brief (PostgreSQL function) iterator on the ipix list of a disk
 *  @param ra right ascension
 *  @param dec declination
 *  @param radius radius of the disk
 *  @param iteration iterator index
 *  @param nside nside level
 *  @return ipix[iteration]
 *
 *  @note
 *  - iterate inside an array of H3C_DEFAULT_MEM_BLOCK elements
 *    the list of ipix is a set of ordered list in which elements follows 
 *    each other
 *    ex: [1,7,9,9,11,19] = {1-7,9,11-19}
 *
 *  @warn H3C_DEFAULT_MEM_BLOCK depends of the value in h3c.sql.in
 */
/*****************************************************************************/
PG_FUNCTION_INFO_V1(pgh3c_radial_query_it);
Datum pgh3c_radial_query_it(PG_FUNCTION_ARGS)
{ 
  static h3c_ipix_t ipix[H3C_DEFAULT_MEM_BLOCK];
  static h3c_coord_t ra_cen_buf,dec_cen_buf, radius_buf;
  static int invocation=0;
  static int nside_buf;
  h3c_coord_t ra_cen = PG_GETARG_FLOAT8(0);
  h3c_coord_t dec_cen = PG_GETARG_FLOAT8(1);
  h3c_coord_t radius = PG_GETARG_FLOAT8(2);
  int iteration = PG_GETARG_INT32(3);
  int nside = PG_GETARG_INT32(4);
  static int max_ipix;


        if (invocation == 0)
        /* If this is the first invocation of the function */
        {
        /* I should set invocation=1 ONLY!!! after setting ra_cen_buf, dec_cen_buf and
         * ipix_buf. Because if the program will be canceled or crashed
         * for some reason the invocation should be == 0
         */
        }
        else
        {
                if ((ra_cen == ra_cen_buf) && (dec_cen == dec_cen_buf) && 
                    (radius == radius_buf) && (nside == nside_buf))
                {
                        if (iteration<max_ipix) PG_RETURN_INT64(ipix[iteration]);
                        PG_RETURN_INT64(-1);
                }
        }

        h3c_debug_init_clock();
        max_ipix = h3c_disk_ipix_it(PG_GETARG_FLOAT8(0),
                                 PG_GETARG_FLOAT8(1),
                                PG_GETARG_FLOAT8(2),
                                nside,
                                ipix,
                                H3C_DEFAULT_MEM_BLOCK) ;

        invocation = 1;
        ra_cen_buf = ra_cen;
        dec_cen_buf = dec_cen;
        radius_buf = radius;
        nside_buf = nside;

        h3c_log(H3C_DEBUG, "pgh3c_radial_query_it end");
        PG_RETURN_INT64(ipix[iteration]);
}

/*****************************************************************************/
/** @brief (PostgreSQL function) iterator on the ipix list of a disk
 *  @param ra right ascension
 *  @param dec declination
 *  @param radius radius of the disk
 *  @param iteration iterator index
 *  @return ipix[iteration]
 *
 *  @note
 *  - default nside=H3C_DEFAULT_NSIDE
 *  - list all the ipix
 *
 *  @see pgh3c_radial_query_it
 */
/*****************************************************************************/
PG_FUNCTION_INFO_V1(pgh3c_radial_query_it_default);
Datum pgh3c_radial_query_it_default(PG_FUNCTION_ARGS)
{ 
  static h3c_ipix_t ipix[H3C_DEFAULT_MEM_BLOCK];
  static h3c_coord_t ra_cen_buf,dec_cen_buf, radius_buf;
  static int invocation=0;
  h3c_coord_t ra_cen = PG_GETARG_FLOAT8(0);
  h3c_coord_t dec_cen = PG_GETARG_FLOAT8(1);
  h3c_coord_t radius = PG_GETARG_FLOAT8(2);
  int iteration = PG_GETARG_INT32(3);
  static int max_ipix;


        if (invocation == 0)
        /* If this is the first invocation of the function */
        {
        /* I should set invocation=1 ONLY!!! after setting ra_cen_buf, dec_cen_buf and
         * ipix_buf. Because if the program will be canceled or crashed
         * for some reason the invocation should be == 0
         */
        }
        else
        {
                if ((ra_cen == ra_cen_buf) && (dec_cen == dec_cen_buf) && 
                    (radius == radius_buf))
                {
                        if (iteration<max_ipix) PG_RETURN_INT64(ipix[iteration]);
                        PG_RETURN_INT64(-1);
                }
        }

        h3c_debug_init_clock();
        max_ipix = h3c_disk_ipix_it(PG_GETARG_FLOAT8(0),
                                 PG_GETARG_FLOAT8(1),
                                PG_GETARG_FLOAT8(2),
                                H3C_DEFAULT_NSIDE,
                                ipix,
                                H3C_DEFAULT_MEM_BLOCK) ;

        invocation = 1;
        ra_cen_buf = ra_cen;
        dec_cen_buf = dec_cen;
        radius_buf = radius;

        h3c_log(H3C_DEBUG, "pgh3c_radial_query_it_default end");
        PG_RETURN_INT64(ipix[iteration]);
}

/*****************************************************************************/
/** @brief (PostgreSQL function) iterator on the ipix list of disks intersection
 *  @param ra right ascension
 *  @param dec declination
 *  @param radius radius of the disk
 *  @param iteration iterator index
 *  @param nside nside level
 *
 *  @note
 *  - Similar than pgh3c_radial_query_it (but best to use the index).
 *    The difference is a smaller number of blocks of ipix.
 *  - iterate inside an array of H3C_DEFAULT_NEARBY_MEM_BLOCK elements
 *    the list of ipix is a set of ordered list in which elements follows 
 *    each other
 *    ex: [1,7,9,9,11,19] = {1-7,9,11-19}
 *
 *  @warn H3C_DEFAULT_NEARBY_MEM_BLOCK depends of the value in h3c.sql.in
 */
/*****************************************************************************/
PG_FUNCTION_INFO_V1(pgh3c_nearby_it);
Datum pgh3c_nearby_it(PG_FUNCTION_ARGS)
{ 
  static h3c_ipix_t ipix[H3C_DEFAULT_NEARBY_MEM_BLOCK];
  static h3c_coord_t ra_cen_buf,dec_cen_buf, radius_buf;
  static int invocation = 0;
  static int nside_buf;
  static int max_ipix;
  h3c_coord_t ra_cen = PG_GETARG_FLOAT8(0);
  h3c_coord_t dec_cen = PG_GETARG_FLOAT8(1);
  h3c_coord_t radius = PG_GETARG_FLOAT8(2);
  int iteration = PG_GETARG_INT32(3);
  int nside = PG_GETARG_INT32(4);
  static int npass = 0;

        if (invocation == 0)
        /* If this is the first invocation of the function */
        {
        /* I should set invocation=1 ONLY!!! after setting ra_cen_buf, dec_cen_buf and
         * ipix_buf. Because if the program will be canceled or crashed
         * for some reason the invocation should be == 0
         */
            h3c_log(H3C_INFO, "initialize pgh3c_nearby_it");
        }
        else
        {
                if (((npass++) % LOG_NPASS)==0) h3c_log(H3C_INFO, "pgh3c_nearby_it_default npass %d", npass);
                if ((ra_cen == ra_cen_buf) && (dec_cen == dec_cen_buf) && 
                    (radius == radius_buf) && (nside == nside_buf))
                {
                        if (iteration < max_ipix) PG_RETURN_INT64(ipix[iteration]);
                        PG_RETURN_INT64(-1);
                }
        }

        h3c_debug_init_clock();
        max_ipix = h3c_disk_ipix_it(PG_GETARG_FLOAT8(0),
                                    PG_GETARG_FLOAT8(1),
                                    PG_GETARG_FLOAT8(2),
                                    nside,
                                    ipix,
                                    H3C_DEFAULT_NEARBY_MEM_BLOCK) ;
        if (max_ipix < 0) h3c_log(H3C_ERROR, "pgh3c_nearby_it_default nothing found");

        invocation = 1;
        ra_cen_buf = ra_cen;
        dec_cen_buf = dec_cen;
        radius_buf = radius;
        nside_buf = nside;

        h3c_log(H3C_DEBUG, "pgh3c_nearby_it_default end");
        PG_RETURN_INT64(ipix[iteration]);
}

/*****************************************************************************/
/** @brief (PostgreSQL function) iterator on the ipix list of disks intersection
 *  @param ra right ascension
 *  @param dec declination
 *  @param radius radius of the disk
 *  @param iteration iterator index
 *
 *  @note
 *  - default nside=H3C_DEFAULT_NSIDE
 *  - list all the ipix
 *
 *  @see pgh3c_nearby_it
 */
/*****************************************************************************/
PG_FUNCTION_INFO_V1(pgh3c_nearby_it_default);
Datum pgh3c_nearby_it_default(PG_FUNCTION_ARGS)
{ 
  static h3c_ipix_t ipix[H3C_DEFAULT_NEARBY_MEM_BLOCK];
  static h3c_coord_t ra_cen_buf,dec_cen_buf, radius_buf;
  static int invocation=0;
  h3c_coord_t ra_cen = PG_GETARG_FLOAT8(0);
  h3c_coord_t dec_cen = PG_GETARG_FLOAT8(1);
  h3c_coord_t radius = PG_GETARG_FLOAT8(2);
  int iteration = PG_GETARG_INT32(3);
  static int max_ipix;
  static int npass = 0;

        if (invocation == 0)
        /* If this is the first invocation of the function */
        {
        /* I should set invocation=1 ONLY!!! after setting ra_cen_buf, dec_cen_buf and
         * ipix_buf. Because if the program will be canceled or crashed
         * for some reason the invocation should be == 0
         */
            h3c_log(H3C_INFO, "initialize pgh3c_nearby_it_default");
        }
        else
        {
                if (((npass++) % LOG_NPASS)==0) h3c_log(H3C_INFO, "pgh3c_nearby_it_default npass %d", npass);
                if ((ra_cen == ra_cen_buf) && (dec_cen == dec_cen_buf) && (radius == radius_buf))
                {
                        if (iteration < max_ipix) PG_RETURN_INT64(ipix[iteration]);
                        PG_RETURN_INT64(-1);
                }
        }

        h3c_debug_init_clock();
        max_ipix = h3c_disk_ipix_it(PG_GETARG_FLOAT8(0),
                                    PG_GETARG_FLOAT8(1),
                                    PG_GETARG_FLOAT8(2),
                                    H3C_DEFAULT_NSIDE,
                                    ipix,
                                    H3C_DEFAULT_NEARBY_MEM_BLOCK) ;
        if (max_ipix < 0) h3c_log(H3C_ERROR, "pgh3c_nearby_it_default nothing found");

        invocation = 1;
        ra_cen_buf = ra_cen;
        dec_cen_buf = dec_cen;
        radius_buf = radius;

        h3c_log(H3C_DEBUG, "pgh3c_nearby_it_default end");
        PG_RETURN_INT64(ipix[iteration]);
}

/*****************************************************************************/
/** @brief (PostgreSQL function) computing the distance between points
 *  @param ra1 right ascension of the first point
 *  @param dec1 declination of the first point
 *  @param ra2 right ascension of the second point
 *  @param dec2 declination of the second point
 *  @return the distance
 */
/*****************************************************************************/
PG_FUNCTION_INFO_V1(pgh3c_dist);
Datum pgh3c_dist(PG_FUNCTION_ARGS)
{
        h3c_coord_t ra1 = PG_GETARG_FLOAT8(0);
        h3c_coord_t dec1 = PG_GETARG_FLOAT8(1);
        h3c_coord_t ra2 = PG_GETARG_FLOAT8(2);
        h3c_coord_t dec2 = PG_GETARG_FLOAT8(3);
        PG_RETURN_FLOAT8(h3c_dist(ra1, dec1, ra2, dec2));
}
 
/*****************************************************************************/
/** @brief (PostgreSQL function) computing the (sinus)distance between points
 *            (to be Q3C compatible)
 *  @param ra1 right ascension of the first point
 *  @param dec1 declination of the first point
 *  @param ra2 right ascension of the second point
 *  @param dec2 declination of the second point
 *  @return the (sinus)distance
 */
/*****************************************************************************/
PG_FUNCTION_INFO_V1(pgh3c_sindist);
Datum pgh3c_sindist(PG_FUNCTION_ARGS)
{
        h3c_coord_t ra1 = PG_GETARG_FLOAT8(0);
        h3c_coord_t dec1 = PG_GETARG_FLOAT8(1);
        h3c_coord_t ra2 = PG_GETARG_FLOAT8(2);
        h3c_coord_t dec2 = PG_GETARG_FLOAT8(3);
        PG_RETURN_FLOAT8(h3c_sindist(ra1, dec1, ra2, dec2));
} 

/*****************************************************************************/
/** @brief memorize polygon and short tests
 *  @param poly_nitems number of points
 *  @param in_ra right ascension memory (IN/OUT)
 *  @param in_dec declination memory (IN/OUT)
 *  @param poly_arr array of points of the polygon
 *  @return 0 new polygon, 1 the same polygon than the previous call
 *
 *  @note get from Q3C library ....
 */
/*****************************************************************************/
static int h3c_array_push(int poly_nitems, 
                        h3c_coord_t *in_ra,
                        h3c_coord_t *in_dec,
                        ArrayType *poly_arr)
{
    int16 typlen;
    bool typbyval;
    char typalign;
    Oid element_type=FLOAT8OID;
    int i, same = 1; /* the same polygon than the previous call */
    char *p;
    h3c_coord_t ra_cur, dec_cur;

#if PG_VERSION_NUM >= 80300
    bits8 *bitmap;
    int bitmask;

        bitmap = ARR_NULLBITMAP(poly_arr);
        bitmask=1;
#endif

        p = ARR_DATA_PTR(poly_arr);
        get_typlenbyvalalign(element_type, &typlen, &typbyval, &typalign);

        for (i = 0; i < poly_nitems; i++)
        {
#if PG_VERSION_NUM >= 80300
            if (bitmap && (*bitmap & bitmask) == 0)
            {
                h3c_log(H3C_ERROR,
                    "null array element not allowed in this context");
            }

            ra_cur = DatumGetFloat8(fetch_att(p, typbyval, typlen));
            p = att_addlength_pointer(p, typlen, PointerGetDatum(p));
            p = (char *) att_align_nominal(p, typalign);
            if (bitmap)
            {
                bitmask <<= 1;
                if (bitmask == 0x100)
                {
                    bitmap++;
                    bitmask = 1;
                }
            }

            if (in_ra[i] != ra_cur) same = 0;
            in_ra[i] = ra_cur;

            if (bitmap && (*bitmap & bitmask) == 0)
            {
                h3c_log(H3C_ERROR,
                    "null array element not allowed in this context");
            }

            dec_cur = DatumGetFloat8(fetch_att(p, typbyval, typlen));
            p = att_addlength_pointer(p, typlen, PointerGetDatum(p));
            p = (char *) att_align_nominal(p, typalign);

            if (bitmap)
            {
                bitmask <<= 1;
                if (bitmask == 0x100)
                {
                    bitmap++;
                    bitmask = 1;
                }
            }

            if (in_dec[i] != dec_cur) same = 0;
            in_dec[i] = dec_cur;

#else
            ra_cur  = DatumGetFloat8(fetch_att(p, typbyval, typlen));
            if (in_ra[i] != ra_cur) same = 0;
            in_ra[i] = ra_cur;

            p = att_addlength(p, typlen, PointerGetDatum(p));
            p = (char *) att_align(p, typalign);
            dec_cur = DatumGetFloat8(fetch_att(p, typbyval, typlen));
            if (in_dec[i] != dec_cur) same = 0;
            in_dec[i] = dec_cur;

            p = att_addlength(p, typlen, PointerGetDatum(p));
            p = (char *) att_align(p, typalign);
#endif

        }

        return same;
}

/*****************************************************************************/
/** @brief (PostgreSQL function) iterator on the ipix list of a polygon
 *  @param polygon array of points
 *  @param iteration iterator index
 *  @param nside nside level
 *  @return ipix[iteration]
 *
 *  @note
 *  - iterate inside an array of H3C_DEFAULT_MEM_BLOCK elements
 *    the list of ipix is a set of ordered list in which elements follows 
 *    each other
 *    ex: [1,7,9,9,11,19] = {1-7,9,11-19}
 */
/*****************************************************************************/
PG_FUNCTION_INFO_V1(pgh3c_poly_query_it);
Datum pgh3c_poly_query_it(PG_FUNCTION_ARGS)
{
        static h3c_coord_t in_ra[H3C_MAX_N_POLY_VERTEX], in_dec[H3C_MAX_N_POLY_VERTEX];
        static h3c_ipix_t ipix[H3C_DEFAULT_MEM_BLOCK];
        static int max_ipix;
        static int nside_buf;

        ArrayType *poly_arr = PG_GETARG_ARRAYTYPE_P(0);
        int iteration = PG_GETARG_UINT32(1);
        int nside = PG_GETARG_UINT32(2);

        int poly_nitems = ArrayGetNItems(ARR_NDIM(poly_arr), ARR_DIMS(poly_arr));
        int same = 1; /* the same polygon than the previous call */

        /* Taken from /pgsql/src/backend/utils/adt/arrayfuncs.c
         * function deconstruct_array
         */

        if (poly_nitems % 2 != 0)
        {
            h3c_log(H3C_ERROR, "Invalid array argument! The array should contain even number of arguments");
        }
        else if (poly_nitems <= 4)
        {
            h3c_log(H3C_ERROR, "Invalid polygon! Less than 3 vertexes");
        }

        poly_nitems /= 2;

        same = h3c_array_push(poly_nitems, in_ra, in_dec, poly_arr);
        if (same && (nside == nside_buf)) {
                if (iteration < max_ipix) PG_RETURN_INT64(ipix[iteration]);
                PG_RETURN_INT64(-1);
        }

        h3c_debug_init_clock();
        max_ipix = h3c_polygon_ipix_it(in_ra, 
                                        in_dec, 
                                        poly_nitems,
                                        nside,
                                        ipix,
                                        H3C_DEFAULT_MEM_BLOCK);
        if (max_ipix < 1) h3c_log(H3C_ERROR, "pgh3c_poly_query_it nothing found");
        nside_buf = nside;

        h3c_log(H3C_DEBUG, "pgh3c_poly_query_it end");
        if (max_ipix > 0) PG_RETURN_INT64(ipix[0]);
        PG_RETURN_INT64(-1);
}

/*****************************************************************************/
/** @brief (PostgreSQL function) iterator on the ipix list of a polygon
 *  @param polygon array of points
 *  @param iteration iterator index
 *  @return ipix[iteration]
 *
 *  @note
 *  - default nside=H3C_DEFAULT_NSIDE
 *
 *  @see pgh3c_radial_query_it_default
 */
/*****************************************************************************/
PG_FUNCTION_INFO_V1(pgh3c_poly_query_it_default);
Datum pgh3c_poly_query_it_default(PG_FUNCTION_ARGS)
{
        static h3c_coord_t in_ra[H3C_MAX_N_POLY_VERTEX], in_dec[H3C_MAX_N_POLY_VERTEX];
          static h3c_ipix_t ipix[H3C_DEFAULT_MEM_BLOCK];
          static int max_ipix;

        ArrayType *poly_arr = PG_GETARG_ARRAYTYPE_P(0);
        int iteration = PG_GETARG_UINT32(1);

        int poly_nitems = ArrayGetNItems(ARR_NDIM(poly_arr), ARR_DIMS(poly_arr));
        int same = 1; /* the same polygon than the previous call */

        /* Taken from /pgsql/src/backend/utils/adt/arrayfuncs.c
         * function deconstruct_array
         */

        if (poly_nitems % 2 != 0)
        {
            h3c_log(H3C_ERROR, "Invalid array argument! The array should contain even number of arguments");
        }
        else if (poly_nitems <= 4)
        {
            h3c_log(H3C_ERROR, "Invalid polygon! Less than 3 vertexes");
        }

        poly_nitems /= 2;

        same = h3c_array_push(poly_nitems, in_ra, in_dec, poly_arr);
        if (same) {
                if (iteration < max_ipix) PG_RETURN_INT64(ipix[iteration]);
                PG_RETURN_INT64(-1);
        }

        h3c_debug_init_clock();
        max_ipix = h3c_polygon_ipix_it(in_ra, 
                                        in_dec, 
                                        poly_nitems,
                                        H3C_DEFAULT_NSIDE,
                                        ipix,
                                        H3C_DEFAULT_MEM_BLOCK);
        if (max_ipix < 0) h3c_log(H3C_ERROR, "pgh3c_poly_query_it_default nothing found");

        h3c_log(H3C_DEBUG, "pgh3c_poly_query_it_default end");
        if (max_ipix > 0) PG_RETURN_INT64(ipix[0]);
        PG_RETURN_INT64(-1);
}

/*****************************************************************************/
/** @brief (PostgreSQL function) get a polygon from a box
 *  @param ra right ascension
 *  @param dec declination
 *  @param width (deg)
 *  @param height(deg)
 *  @return an array {ra0,de0,ra1,de1,....}
*/
/*****************************************************************************/
PG_FUNCTION_INFO_V1(pgh3c_box_to_poly);
Datum pgh3c_box_to_poly(PG_FUNCTION_ARGS)
{
        static ArrayType  *result;
        Datum  *data;
        int16       typlen;
        bool        typbyval;
        char        typalign;

        static h3c_coord_t in_ra[4];
        static h3c_coord_t in_dec[4];
        static h3c_coord_t ra_cen_buf, dec_cen_buf;
        static h3c_coord_t width_buf, height_buf;

        if (PG_GETARG_FLOAT8(0) == ra_cen_buf  && 
            PG_GETARG_FLOAT8(1) == dec_cen_buf &&
            PG_GETARG_FLOAT8(2) == width_buf &&
            PG_GETARG_FLOAT8(3) == height_buf) {

        }
        else {
            ra_cen_buf  = PG_GETARG_FLOAT8(0);
            dec_cen_buf = PG_GETARG_FLOAT8(1);
            width_buf   = PG_GETARG_FLOAT8(2);
            height_buf  = PG_GETARG_FLOAT8(3);
    
            h3c_box_2_polygon(PG_GETARG_FLOAT8(0),
                              PG_GETARG_FLOAT8(1),
                              PG_GETARG_FLOAT8(2),
                              PG_GETARG_FLOAT8(3),
                              in_ra, in_dec);
        }
 
        data = ( Datum *) palloc(sizeof(Datum)*8);
        data[0] = Float8GetDatum (in_ra[0]);
        data[1] = Float8GetDatum (in_dec[0]);
        data[2] = Float8GetDatum (in_ra[1]);
        data[3] = Float8GetDatum (in_dec[1]);
        data[4] = Float8GetDatum (in_ra[2]);
        data[5] = Float8GetDatum (in_dec[2]);
        data[6] = Float8GetDatum (in_ra[3]);
        data[7] = Float8GetDatum (in_dec[3]);

        /* get required info about the element type */
        get_typlenbyvalalign(FLOAT8OID, &typlen, &typbyval, &typalign);

        /* now build the array */
        result = construct_array(data, 8, FLOAT8OID, typlen, typbyval, typalign);

        PG_RETURN_ARRAYTYPE_P(result);
}

#if 0
/*****************************************************************************/
/** @brief (PostgreSQL function) iterator on the ipix list of a box
 *  @param ra right ascension
 *  @param dec declination
 *  @param width (deg)
 *  @param height(deg)
 *  @param iteration iterator index
 *  @param nside nside level
 *  @return ipix[iteration]
 *
 *  @note
 *  - iterate inside an array of H3C_DEFAULT_MEM_BLOCK elements
 *    the list of ipix is a set of ordered list in which elements follows 
 *    each other
 *    ex: [1,7,9,9,11,19] = {1-7,9,11-19}
 */
/*****************************************************************************/
PG_FUNCTION_INFO_V1(pgh3c_box_query_it);
Datum pgh3c_box_query_it(PG_FUNCTION_ARGS)
{
        static int invocation=0;
        static h3c_coord_t ra_cen_buf,dec_cen_buf, height_buf, width_buf;
        static h3c_ipix_t ipix[H3C_DEFAULT_MEM_BLOCK];
        static int max_ipix;
        static int nside_buf;
        h3c_coord_t in_ra[4], in_dec[4];

        h3c_coord_t ra_cen = PG_GETARG_FLOAT8(0);
        h3c_coord_t dec_cen = PG_GETARG_FLOAT8(1);
        h3c_coord_t width = PG_GETARG_FLOAT8(2);
        h3c_coord_t height = PG_GETARG_FLOAT8(3);
        
        int iteration = PG_GETARG_UINT32(4);
        int nside = PG_GETARG_UINT32(5);


        if (invocation == 0)
        /* If this is the first invocation of the function */
        {
        /* I should set invocation=1 ONLY!!! after setting ra_cen_buf, dec_cen_buf and
         * ipix_buf. Because if the program will be canceled or crashed
         * for some reason the invocation should be == 0
         */
            h3c_log(H3C_INFO, "initialize pgh3c_box_query_it");
        }
        else
        {
                if ((ra_cen == ra_cen_buf) && (dec_cen == dec_cen_buf) 
                 && (height == height_buf) && (width == width_buf)
                 && nside == nside_buf)
                {
                        if (iteration < max_ipix) PG_RETURN_INT64(ipix[iteration]);
                        PG_RETURN_INT64(-1);
                }
        }

        _box_2_polygon(ra_cen, dec_cen, width, height, in_ra, in_de);
        
        h3c_debug_init_clock();
        max_ipix = h3c_polygon_ipix_it(in_ra, 
                                       in_dec, 
                                       4,
                                       nside,
                                       ipix,
                                       H3C_DEFAULT_MEM_BLOCK);
        if (max_ipix < 1) h3c_log(H3C_ERROR, "pgh3c_box_query_it nothing found");

        nside_buf = nside;
        invocation = 1;
        ra_cen_buf = ra_cen;
        dec_cen_buf = dec_cen;
        width_buf = width;
        height_buf = height;

        h3c_log(H3C_DEBUG, "pgh3c_box_query_it end");
        if (max_ipix > 0) PG_RETURN_INT64(ipix[0]);
        PG_RETURN_INT64(-1);
}

/*****************************************************************************/
/** @brief (PostgreSQL function) iterator on the ipix list of a box
 *  @param ra right ascension
 *  @param dec declination
 *  @param width (deg)
 *  @param height(deg)
 *  @param iteration iterator index
 *  @return ipix[iteration]
 *
 *  @note
 *  - default nside=H3C_DEFAULT_NSIDE
 *
 *  @see pgh3c_box_query_it_default
 */
/*****************************************************************************/
PG_FUNCTION_INFO_V1(pgh3c_box_query_it_default);
Datum pgh3c_box_query_it_default(PG_FUNCTION_ARGS)
{
        static int invocation=0;
        static h3c_coord_t ra_cen_buf,dec_cen_buf, height_buf, width_buf;
        static h3c_ipix_t ipix[H3C_DEFAULT_MEM_BLOCK];
        static int max_ipix;
        h3c_coord_t in_ra[4], in_dec[4];

        h3c_coord_t ra_cen = PG_GETARG_FLOAT8(0);
        h3c_coord_t dec_cen = PG_GETARG_FLOAT8(1);
        h3c_coord_t width = PG_GETARG_FLOAT8(2);
        h3c_coord_t height = PG_GETARG_FLOAT8(3);
        
        int iteration = PG_GETARG_UINT32(4);

        if (invocation == 0)
        /* If this is the first invocation of the function */
        {
        /* I should set invocation=1 ONLY!!! after setting ra_cen_buf, dec_cen_buf and
         * ipix_buf. Because if the program will be canceled or crashed
         * for some reason the invocation should be == 0
         */
            h3c_log(H3C_INFO, "initialize pgh3c_box_query_it");
        }
        else
        {
                if ((ra_cen == ra_cen_buf) && (dec_cen == dec_cen_buf) 
                 && (height == height_buf) && (width == width_buf))
                {
                        if (iteration < max_ipix) PG_RETURN_INT64(ipix[iteration]);
                        PG_RETURN_INT64(-1);
                }
        }

        _box_2_polygon(ra_cen, dec_cen, width, height, in_ra, in_de);

        if (iteration ==0) 
            h3c_log(H3C_DEBUG, "pgh3c_box_query_it: update to polygon h3c_poly_query: (%f,%f),(%f,%f), (%f,%f), (%f,%f)",
                    in_ra[0], in_dec[0], in_ra[1], in_dec[1], in_ra[2], in_dec[2], in_ra[3], in_dec[3]);

        h3c_debug_init_clock();
        max_ipix = h3c_polygon_ipix_it(in_ra, 
                                       in_dec, 
                                       4,
                                       H3C_DEFAULT_NSIDE,
                                       ipix,
                                       H3C_DEFAULT_MEM_BLOCK);
        if (max_ipix < 1) h3c_log(H3C_ERROR, "pgh3c_box_query_it nothing found");

        invocation = 1;
        ra_cen_buf = ra_cen;
        dec_cen_buf = dec_cen;
        width_buf = width;
        height_buf = height;

        h3c_log(H3C_DEBUG, "pgh3c_box_query_it end");
        if (max_ipix > 0) PG_RETURN_INT64(ipix[0]);
        PG_RETURN_INT64(-1);
}
#endif
/*****************************************************************************/
/** @brief (PostgreSQL function) point inside the polygon
 *  @param ra_cen right ascention of the point
 *  @param dev_cen declination of the point
 *  @param poly the list of points (of the polygon)
 *  @return true/false
 *
 *  @warn the convex polygons are just a particular case of this ensemble
 *
 *  This function can be applied for polygon more large than convex.
 *  (see h3c_poly_is_convex)
 *
 *  @note 
 *  - take care of the sens (ie: the order of the point)
 *  - algo from F.Ochsenbein
 */
/*****************************************************************************/
PG_FUNCTION_INFO_V1(pgh3c_in_poly_convex);
Datum pgh3c_in_poly_convex(PG_FUNCTION_ARGS)
{
        static h3c_coord_t in_ra[H3C_MAX_N_POLY_VERTEX], in_dec[H3C_MAX_N_POLY_VERTEX];

        ArrayType *poly_arr = PG_GETARG_ARRAYTYPE_P(2);
        h3c_coord_t ra_cen = PG_GETARG_FLOAT8(0);
        h3c_coord_t dec_cen = PG_GETARG_FLOAT8(1);

        int poly_nitems = ArrayGetNItems(ARR_NDIM(poly_arr), ARR_DIMS(poly_arr));
        static bool result;

         /*h3c_debug_init_clock();*/

        /* Taken from /pgsql/src/backend/utils/adt/arrayfuncs.c
         * function deconstruct_array
         */

        if (poly_nitems % 2 != 0)
        {
            h3c_log(H3C_ERROR, "Invalid array argument! The array should contain even number of arguments");
        }
        else if (poly_nitems <= 4)
        {
            h3c_log(H3C_ERROR, "Invalid polygon! Less than 3 vertexes");
        }

        poly_nitems /= 2;

        h3c_array_push(poly_nitems, in_ra, in_dec, poly_arr);
        if ( !h3c_poly_is_convex(poly_nitems, in_ra, in_dec)) {
            h3c_log(H3C_ERROR, "The polygon is not convex");
        }

        result = (h3c_check_sphere_point_in_poly_convex(poly_nitems, in_ra, in_dec,
                                                        ra_cen, dec_cen)) != H3C_DISJUNCT;

        /*h3c_debug("pgh3c_in_poly_convex end");*/
        PG_RETURN_BOOL((result));
}

/*****************************************************************************/
/** @brief (PostgreSQL function) point inside the polygon
 *  @param ra_cen right ascention of the point
 *  @param dev_cen declination of the point
 *  @param poly the list of points (of the polygon)
 *  @return true/false
 *
 *  @warn some bugs still exist:
 *
 *  ex: (45.0,31.38816) in {50,30, 10,0 , 0,10, 0,55}
 *      (55.37, 37.26)  in {50,30, 10,0 , 0,10, 0,55}
 */
/*****************************************************************************/
PG_FUNCTION_INFO_V1(pgh3c_in_poly);
Datum pgh3c_in_poly(PG_FUNCTION_ARGS)
{
        static h3c_coord_t in_ra[H3C_MAX_N_POLY_VERTEX], in_dec[H3C_MAX_N_POLY_VERTEX];

        ArrayType *poly_arr = PG_GETARG_ARRAYTYPE_P(2);
        h3c_coord_t ra_cen = PG_GETARG_FLOAT8(0);
        h3c_coord_t dec_cen = PG_GETARG_FLOAT8(1);

        int poly_nitems = ArrayGetNItems(ARR_NDIM(poly_arr), ARR_DIMS(poly_arr));
        static bool result;

         /*h3c_debug_init_clock();*/

        /* Taken from /pgsql/src/backend/utils/adt/arrayfuncs.c
         * function deconstruct_array
         */

        if (poly_nitems % 2 != 0)
        {
            h3c_log(H3C_ERROR, "Invalid array argument! The array should contain even number of arguments");
        }
        else if (poly_nitems <= 4)
        {
            h3c_log(H3C_ERROR, "Invalid polygon! Less than 3 vertexes");
        }

        poly_nitems /= 2;

        h3c_array_push(poly_nitems, in_ra, in_dec, poly_arr);

        result = (h3c_check_sphere_point_in_poly(poly_nitems, in_ra, in_dec,
                                                 ra_cen, dec_cen)) != H3C_DISJUNCT;

        /*h3c_debug("pgh3c_in_poly end");*/
        PG_RETURN_BOOL((result));
}

#if 0
/*****************************************************************************/
/** @brief (PostgreSQL function) point inside the box
 *  @param ra_cen  right ascention of the point
 *  @param dec_cen declination of the point
 *  @param width   RA length of the box (deg)
 *  @param height  DEC length of the box (deg)
 *  @return true/false
 */
/*****************************************************************************/
PG_FUNCTION_INFO_V1(pgh3c_in_box);
Datum pgh3c_in_box(PG_FUNCTION_ARGS)
{
    h3c_coord_t in_ra[4], in_dec[4];
    h3c_coord_t ra_cen = PG_GETARG_FLOAT8(0);
    h3c_coord_t dec_cen = PG_GETARG_FLOAT8(1);
    h3c_coord_t width = PG_GETARG_FLOAT8(2);
    h3c_coord_t height = PG_GETARG_FLOAT8(3);

    _box_2_polygon(ra_cen, dec_cen, width, height, in_ra, in_dec);

    result = (h3c_check_sphere_point_in_poly(poly_nitems, in_ra, in_dec,
                                             ra_cen, dec_cen)) != H3C_DISJUNCT;

    /*h3c_debug("pgh3c_in_poly end");*/
    PG_RETURN_BOOL((result));
}
#endif

/*****************************************************************************/
/** @brief (PostgreSQL function) centre of a polygon [added]
 *  @param poly list of points of the polygon
 *  @return the center (ra,dec)
 */
/*****************************************************************************/
PG_FUNCTION_INFO_V1(pgh3c_poly_center);
Datum pgh3c_poly_center(PG_FUNCTION_ARGS)
{        
        static h3c_coord_t in_ra[H3C_MAX_N_POLY_VERTEX], in_dec[H3C_MAX_N_POLY_VERTEX];
        ArrayType *poly_arr = PG_GETARG_ARRAYTYPE_P(0);
        int poly_nitems = ArrayGetNItems(ARR_NDIM(poly_arr), ARR_DIMS(poly_arr));
        h3c_coord_t radec[2];
        Datum      *data;
        int16       typlen;
        bool        typbyval;
        char        typalign;
        ArrayType  *result;

        if (poly_nitems % 2 != 0)
        {
            h3c_log(H3C_ERROR, "Invalid array argument! The array should contain even number of arguments");
        }
        else if (poly_nitems <= 4)
        {
            h3c_log(H3C_ERROR, "Invalid polygon! Less than 3 vertexes");
        }

        poly_nitems /= 2;

        h3c_array_push(poly_nitems, in_ra, in_dec, poly_arr);
        if ( !h3c_poly_is_convex(poly_nitems, in_ra, in_dec)) {
            h3c_log(H3C_ERROR, "The polygon is not convex");
        }

        h3c_poly_center(poly_nitems, in_ra, in_dec, radec);

        data = ( Datum *) palloc(sizeof(Datum)*2);
        data[0] = Float8GetDatum (radec[0]);
        data[1] = Float8GetDatum (radec[1]);
        /* get required info about the element type */
        get_typlenbyvalalign(FLOAT8OID, &typlen, &typbyval, &typalign);

        /* now build the array */
        result = construct_array(data, 2, FLOAT8OID, typlen, typbyval, typalign);

        PG_RETURN_ARRAYTYPE_P(result);
}

/*****************************************************************************/
/** @brief (PostgreSQL function) if a sphere intersect a polygon [added]
 *  @param ra right ascention
 *  @param dec declination
 *  @param radius radius of the circle
 *  @param polygon array of points
 */
/*****************************************************************************/
PG_FUNCTION_INFO_V1(pgh3c_circle_intersect_poly);
Datum pgh3c_circle_intersect_poly(PG_FUNCTION_ARGS)
{
        static h3c_coord_t in_ra[H3C_MAX_N_POLY_VERTEX], in_dec[H3C_MAX_N_POLY_VERTEX];
        static bool result;
        int poly_nitems;
        ArrayType *poly_arr;

        h3c_coord_t ra_cen = PG_GETARG_FLOAT8(0);
        h3c_coord_t dec_cen = PG_GETARG_FLOAT8(1);
        h3c_coord_t radius = PG_GETARG_FLOAT8(2);
        poly_arr = PG_GETARG_ARRAYTYPE_P(3);

        poly_nitems = ArrayGetNItems(ARR_NDIM(poly_arr), ARR_DIMS(poly_arr));
        if (poly_nitems % 2 != 0)
        {
            h3c_log(H3C_ERROR, "Invalid array argument! The array should contain even number of arguments");
        }
        else if (poly_nitems <= 4)
        {
            h3c_log(H3C_ERROR, "Invalid polygon! Less than 3 vertexes");
        }
        poly_nitems /= 2;

        h3c_array_push(poly_nitems, in_ra, in_dec, poly_arr);
        if ( !h3c_poly_is_convex(poly_nitems, in_ra, in_dec)) {
            h3c_log(H3C_ERROR, "The polygon is not convex");
        }

        result = (h3c_check_sphere_circle_intersect_poly(poly_nitems, in_ra, in_dec,
                                                         ra_cen, dec_cen, radius)) != H3C_DISJUNCT;

        PG_RETURN_BOOL((result));
}

/*****************************************************************************/
/** @brief intersection between polygon
 *  @param polygon 1
 *  @param polygon 2
 *  @return boolean
 */
/*****************************************************************************/
PG_FUNCTION_INFO_V1(pgh3c_poly_intersect_poly);
Datum pgh3c_poly_intersect_poly(PG_FUNCTION_ARGS)
{
        static h3c_coord_t in_ra1[H3C_MAX_N_POLY_VERTEX], in_dec1[H3C_MAX_N_POLY_VERTEX];
        static h3c_coord_t in_ra2[H3C_MAX_N_POLY_VERTEX], in_dec2[H3C_MAX_N_POLY_VERTEX];
        static bool result;
        int poly_nitems1, poly_nitems2;
        ArrayType *poly_arr1, *poly_arr2;

        poly_arr1 = PG_GETARG_ARRAYTYPE_P(0);
        poly_arr2 = PG_GETARG_ARRAYTYPE_P(1);

        poly_nitems1 = ArrayGetNItems(ARR_NDIM(poly_arr1), ARR_DIMS(poly_arr1));
        if (poly_nitems1 % 2 != 0)
        {
            h3c_log(H3C_ERROR, "Invalid array argument! The array should contain even number of arguments");
        }
        else if (poly_nitems1 <= 4)
        {
            h3c_log(H3C_ERROR, "Invalid polygon! Less than 3 vertexes");
        }
        poly_nitems1 /= 2;

        h3c_array_push(poly_nitems1, in_ra1, in_dec1, poly_arr1);
        if ( !h3c_poly_is_convex(poly_nitems1, in_ra1, in_dec1)) {
            h3c_log(H3C_ERROR, "The polygon1 is not convex");
        }

        poly_nitems2 = ArrayGetNItems(ARR_NDIM(poly_arr2), ARR_DIMS(poly_arr2));
        if (poly_nitems2 % 2 != 0)
        {
            h3c_log(H3C_ERROR, "Invalid array argument! The array should contain even number of arguments");
        }
        else if (poly_nitems2 <= 4)
        {
            h3c_log(H3C_ERROR, "Invalid polygon! Less than 3 vertexes");
        }
        poly_nitems2 /= 2;

        h3c_array_push(poly_nitems2, in_ra2, in_dec2, poly_arr2);
        if ( !h3c_poly_is_convex(poly_nitems2, in_ra2, in_dec2)) {
            h3c_log(H3C_ERROR, "The polygon2 is not convex");
        }

        result = (h3c_check_poly_intersect_poly(poly_nitems1, in_ra1, in_dec1,
                                                poly_nitems2, in_ra2, in_dec2)) != H3C_DISJUNCT;

        PG_RETURN_BOOL((result));
}

/*****************************************************************************/
/** @brief (PostgreSQL function) distance between 2 ipix center [added]
 *  @param ipix1 first ipix 
 *  @param ipix2 second ipix
 *  @return the distance in deg
 */
/*****************************************************************************/
PG_FUNCTION_INFO_V1(pgh3c_dist_ipix);
Datum pgh3c_dist_ipix(PG_FUNCTION_ARGS)
{
        h3c_coord_t coo1[2], coo2[2], d;
        h3c_ipix_t ipix1 = PG_GETARG_INT64(0);
        h3c_ipix_t ipix2 = PG_GETARG_INT64(1);

        h3c_pix2ang_nest(H3C_DEFAULT_NSIDE,
                     ipix1,
                     &coo1[1],
                     &coo1[0]);
        coo1[1]=h3c_theta2dec(coo1[1])*H3C_RADEG;
        coo1[0]=h3c_phi2ra(coo1[0])*H3C_RADEG;

        h3c_pix2ang_nest(H3C_DEFAULT_NSIDE,
                     ipix2,
                     &coo2[1],
                     &coo2[0]);
        coo2[1]=h3c_theta2dec(coo2[1])*H3C_RADEG;
        coo2[0]=h3c_phi2ra(coo2[0])*H3C_RADEG;

        d = h3c_dist(coo1[0], coo1[1], coo2[0], coo2[2]);
        PG_RETURN_FLOAT8(d);
}

/*****************************************************************************/
/** @brief (PostgreSQL function) verify if a point is in an ellipse
 *  @param ra right ascention of the point to verify
 *  @param dec declination of tthe point to verify
 *  @param ra_cen right ascention of the center
 *  @param dec_cen declination of the center
 *  @param radius major axis in deg
 *  @param axis_ratio ratio between major and minor axis
 *  @param PA position angle
 *  @return boolean
 */
/*****************************************************************************/
PG_FUNCTION_INFO_V1(pgh3c_in_ellipse);
Datum pgh3c_in_ellipse(PG_FUNCTION_ARGS)
{

        h3c_coord_t ra = PG_GETARG_FLOAT8(0); /* ra_cen */
        h3c_coord_t dec = PG_GETARG_FLOAT8(1); /* dec_cen */
        h3c_coord_t ra_cen = PG_GETARG_FLOAT8(2); /* ra_cen */
        h3c_coord_t dec_cen = PG_GETARG_FLOAT8(3); /* dec_cen */
        h3c_coord_t radius = PG_GETARG_FLOAT8(4); /* error radius */
        h3c_coord_t axis_ratio = PG_GETARG_FLOAT8(5); /* axis_ratio */
        h3c_coord_t PA = PG_GETARG_FLOAT8(6); /* PA */
        h3c_coord_t e = sqrt(1 - axis_ratio * axis_ratio);
        bool result = h3c_in_ellipse(ra_cen, dec_cen, ra,dec, radius, e, PA);
        PG_RETURN_BOOL(result);
}

/*****************************************************************************/
/** @brief (PostgreSQL function) iterator on the ipix list of an ellipse
 *  @param ra_cen right ascension of the ellipse center
 *  @param dec_cen declination of the ellipse center
 *  @param radius radius of the ellipse (or major axis)
 *  @param axis_ratio ratio between minor and major axis
 *  @param iterator iterator index
 *  @param nside the nside level
 */
/*****************************************************************************/
PG_FUNCTION_INFO_V1(pgh3c_ellipse_query_it);
Datum pgh3c_ellipse_query_it(PG_FUNCTION_ARGS)
{ 
  static h3c_ipix_t ipix[H3C_DEFAULT_MEM_BLOCK];
  static h3c_coord_t ra_cen_buf,dec_cen_buf, radius_buf;
  static int invocation=0;
  static int nside_buf, axis_ratio_buf;
  h3c_coord_t ra_cen = PG_GETARG_FLOAT8(0);
  h3c_coord_t dec_cen = PG_GETARG_FLOAT8(1);
  h3c_coord_t radius = PG_GETARG_FLOAT8(2);
  h3c_coord_t axis_ratio = PG_GETARG_FLOAT8(3);
  int iteration = PG_GETARG_INT32(4);
  int nside = PG_GETARG_INT32(5);
  static int max_ipix;

        if (invocation == 0)
        /* If this is the first invocation of the function */
        {
        /* I should set invocation=1 ONLY!!! after setting ra_cen_buf, dec_cen_buf and
         * ipix_buf. Because if the program will be canceled or crashed
         * for some reason the invocation should be == 0
         */
        }
        else
        {
                if ((ra_cen == ra_cen_buf) && (dec_cen == dec_cen_buf) && 
                    (radius == radius_buf) && (nside == nside_buf) && 
                    (axis_ratio == axis_ratio_buf))
                {
                        if (iteration<max_ipix) PG_RETURN_INT64(ipix[iteration]);
                        PG_RETURN_INT64(-1);
                }
        }

        h3c_debug_init_clock();
        max_ipix = h3c_disk_ipix_it(PG_GETARG_FLOAT8(0),
                                 PG_GETARG_FLOAT8(1),
                                PG_GETARG_FLOAT8(2)/2.,
                                nside,
                                ipix,
                                H3C_DEFAULT_MEM_BLOCK) ;
        if (max_ipix < 0) h3c_log(H3C_ERROR, "pgh3c_ellipse_query_it nothing found");

        invocation = 1;
        ra_cen_buf = ra_cen;
        dec_cen_buf = dec_cen;
        radius_buf = radius;
        nside_buf = nside;
        axis_ratio_buf = axis_ratio;

        h3c_log(H3C_DEBUG, "pgh3c_ellipse_query_it end");
        PG_RETURN_INT64(ipix[iteration]);
}

/*****************************************************************************/
/** @brief (PostgreSQL function) calculate the area of a disk
 *  @param radius radius of the circle
 *  @return the area
 */
/*****************************************************************************/
PG_FUNCTION_INFO_V1(pgh3c_circle_area);
Datum pgh3c_circle_area(PG_FUNCTION_ARGS)
{ 
        h3c_coord_t radius = PG_GETARG_FLOAT8(0);
        h3c_coord_t val = sin(radius*H3C_DEGRA/2.);
        PG_RETURN_FLOAT8(4*H3C_PI*val*val);
}

/*****************************************************************************/
/** @brief (PostgreSQL function) 
 *  @param poly array {ra0,dec0, ra1,dec1, ra2,dec2,....}
 *  @return the area (deg x deg)
 */
/*****************************************************************************/
PG_FUNCTION_INFO_V1(pgh3c_poly_area);
Datum pgh3c_poly_area(PG_FUNCTION_ARGS)
{ 
        h3c_coord_t area;
        static h3c_coord_t in_ra[H3C_MAX_N_POLY_VERTEX], in_dec[H3C_MAX_N_POLY_VERTEX];
        ArrayType *poly_arr = PG_GETARG_ARRAYTYPE_P(0);
        int poly_nitems = ArrayGetNItems(ARR_NDIM(poly_arr), ARR_DIMS(poly_arr));

        /* Taken from /pgsql/src/backend/utils/adt/arrayfuncs.c
         * function deconstruct_array
         */

        if (poly_nitems % 2 != 0)
        {
            h3c_log(H3C_ERROR, "Invalid array argument! The array should contain even number of arguments");
        }
        else if (poly_nitems <= 4)
        {
            h3c_log(H3C_ERROR, "Invalid polygon! Less than 3 vertexes");
        }

        poly_nitems /= 2;
        h3c_array_push(poly_nitems, in_ra, in_dec, poly_arr);

        h3c_debug_init_clock();
        if ( !h3c_poly_is_convex(poly_nitems, in_ra, in_dec)) {
            h3c_log(H3C_ERROR, "The polygon is not convex");
        }
        area = h3c_polygon_area(in_ra, in_dec, poly_nitems);

        h3c_log(H3C_DEBUG, "pgh3c_poly_area end");
        PG_RETURN_FLOAT8(area);
}

/*****************************************************************************/
/** @brief get HEALPix in default nside (PostgreSQL function) 
 *  @param HEALPix number
 *  @param nside (INPUT)
 *  @iteration [0=min,1=max]
 *  @return  the HEALPix nuber in the default nside
 */
/*****************************************************************************/
PG_FUNCTION_INFO_V1(pgh3c_ipix2default_it);
Datum pgh3c_ipix2default_it(PG_FUNCTION_ARGS)
{ 
  static h3c_ipix_t m_ipix[2];
  static h3c_ipix_t ipix_buf, nside_buf;
  static int invocation = 0;
  h3c_ipix_t ipix = PG_GETARG_INT32(0);
  int nside = PG_GETARG_INT32(1);
  int iteration = PG_GETARG_INT32(2);
  int order, order_default;

        if (invocation == 0)
        /* If this is the first invocation of the function */
        {
        /* I should set invocation=1 ONLY!!! after setting ra_cen_buf, dec_cen_buf and
         * ipix_buf. Because if the program will be canceled or crashed
         * for some reason the invocation should be == 0
         */
        }
        else
        {
                if ((ipix == ipix_buf) && (nside == nside_buf))
                {
                        if (iteration < 2) PG_RETURN_INT64(m_ipix[iteration]);
                        PG_RETURN_INT64(-1);
                }
        }

        order_default = h3c_order(H3C_DEFAULT_NSIDE);
        order = h3c_order(nside);
        if (order == order_default) {
            m_ipix[0] = ipix; 
            m_ipix[1] = ipix+1;
        }
        else if (order < order_default) {
            m_ipix[0] = ipix<<(order_default-order)*2;
            m_ipix[1] = (ipix+1)<<(order_default-order)*2;
        }
        else {
            m_ipix[0] = ipix>>(order_default-order)*2;
            m_ipix[1] = (ipix+1)>>(order_default-order)*2;
        }

        invocation = 1;
        ipix_buf = ipix;
        nside_buf = nside;

        h3c_log(H3C_DEBUG, "pgh3c_ipix2default_it");
        PG_RETURN_INT64(m_ipix[iteration]);
}

/*****************************************************************************/
/** @brief get NSIDE of the configuration indexation 
 *  @return  nside
 */
/*****************************************************************************/
PG_FUNCTION_INFO_V1(pgh3c_nside_config);
Datum pgh3c_nside_config(PG_FUNCTION_ARGS)
{
        PG_RETURN_INT64(H3C_DEFAULT_NSIDE);
}


