/*
	   Copyright (C) 2012 Gilles Landais (CDS)

    Author: Gilles Landais, Strasbourg astronomical Data Center (CDS)
    Email: gilles.landais@unistra.fr

    This file is part of H3C.

    H3C is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    H3C is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with H3C; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

/** @file h3cpp_healpix.cc
 *  @brief interface the Healpix C++ library
 *  @author G.Landais (CDS)
 *  @date 26/07/2011
 */

/* Local Includes */
#include "healpix_base.h"
#include "common.h"
#include "h3c_util.h"
#include "h3cpp_healpix.h"


/*****************************************************************************/
/** @brief retrieve the ipix list for a cone
 *  @param ra right ascension
 *  @param dec declination
 *  @param radius radius
 *  @param nside nside level
 *  @param nresult number of elements in result
 *  @return ipix list 
 *          NULL if no ipix found
 *               if too much ipix (*nresult > 0)
 */
/*****************************************************************************/
h3c_ipix_t *h3cpp_disk_ipix(h3c_coord_t ra,
                            h3c_coord_t dec,
                            double radius,
                            int nside,
                            int *nresult, 
                            int inclusive) 
{
    int i,order;
    pointing ptg;
#ifdef H3C_INT8
    vector<int64> list;
#else
    vector<int> list;
#endif

    h3c_ipix_t *retlist = NULL;
    *nresult = 0;

    ptg.theta = h3c_dec2theta(dec*H3C_DEGRA);
    ptg.phi = h3c_ra2phi(ra*H3C_DEGRA);
    order = h3c_order(nside);

#ifdef H3C_INT8
#if 0
    T_Healpix_Base<int64> healpix(order, NEST);
#else
    T_Healpix_Base<int64> healpix(order, RING);
#endif
#else
#if 0
    Healpix_Base healpix(order, NEST);
#else
    Healpix_Base healpix(order, RING);
#endif
#endif

    if (inclusive)
        healpix.query_disc_inclusive(ptg,radius*H3C_DEGRA,list);
    else
        healpix.query_disc(ptg,radius*H3C_DEGRA,list);

    if (list.size() > 0 && list.size() < H3C_DEFAULT_MAX_IPIX) {
        retlist = (h3c_ipix_t*)malloc(sizeof(h3c_ipix_t)*(list.size()));
        for (i = 0; i < list.size(); i++)
#if 0
            retlist[i] = list[i];
#else
            retlist[i] = healpix.ring2nest(list[i]);
#endif
    }
    else h3c_log(H3C_INFO, "too much ipix %d>%d", list.size(), H3C_DEFAULT_MAX_IPIX);

    *nresult = list.size();
    list.clear();
    return retlist;
}

/*****************************************************************************/
/** @brief get the number of ipix for a given nside
 *  @param nside the nside level
 *  @return number of ipix
 */
/*****************************************************************************/
int h3cpp_number_ipix(int nside) {
    int order = h3c_order(nside);

#ifdef H3C_INT8
    T_Healpix_Base<int64> healpix(order, RING);
#else
    Healpix_Base healpix(order, RING);
#endif

    return  healpix.Npix();
}

/*****************************************************************************/
/** @brief get the max radius of the circles arround the ipix for a given nside 
 *  @param nside the nside level
 *  @return the radius
 */
/*****************************************************************************/
h3c_coord_t h3cpp_radius(int nside) {
    int order = h3c_order(nside);

#ifdef H3C_INT8
    T_Healpix_Base<int64> healpix(order, RING);
#else
    Healpix_Base healpix(order, RING);
#endif
    return healpix.max_pixrad()*H3C_RADEG;
}

/*****************************************************************************/
/** @brief get ipix neighbors
 *  @param ipix the ipix
 *  @param nside the nside level
 *  @return Array containing the ipix (static)
 */
/*****************************************************************************/
h3c_ipix_t *h3cpp_get_neighbors(h3c_ipix_t ipix, int nside) {
    static h3c_ipix_t neighbors[8];
    int i,order = h3c_order(nside);

#ifdef H3C_INT8
    T_Healpix_Base<int64> healpix(order, NEST);
    fix_arr < int64, 8 > result;
#else
    Healpix_Base healpix(order, NEST);
    fix_arr < int, 8 > result;
#endif
    healpix.neighbors(ipix, result);
    for (i = 0; i < 8; i++) neighbors[i] = result[i];
    return neighbors;
}

/*****************************************************************************/
/* @brief pixel (in NEST) to angle (theta,phi)
 * @param nside the nside level
 * @param ipnest ipix (NEST)
 * @param theta angle (OUT))
 * @param phi angle   (OUT)
 */
/*****************************************************************************/
void h3cpp_pix2ang_nest(long nside,
                        h3c_ipix_t ipnest,
                        h3c_coord_t *theta,
                        h3c_coord_t *phi){
    int order = h3c_order(nside);
#ifdef H3C_INT8
    T_Healpix_Base<int64> healpix(order, NEST);
#else
    Healpix_Base healpix(order, NEST);
#endif
    pointing ptg;

    ptg = healpix.pix2ang(ipnest);
    *theta = ptg.theta;
    *phi = ptg.phi;
}

/*****************************************************************************/
/* @brief angle (theta,phi) to ipix (NEST)
 * @param nside the nside level
 * @param theta angle (OUT))
 * @param phi angle   (OUT)
 * @param ipnest ipix (NEST)
 */
/*****************************************************************************/
void h3cpp_ang2pix_nest(long nside,
                        h3c_coord_t theta,
                        h3c_coord_t phi,
                        h3c_ipix_t *ipnest){
    int order = h3c_order(nside);
#ifdef H3C_INT8
    T_Healpix_Base<int64> healpix(order, NEST);
#else
    Healpix_Base healpix(order, NEST);
#endif
    pointing ptg;
    ptg.theta = theta;
    ptg.phi = phi;

    *ipnest = healpix.ang2pix(ptg);
}

/*****************************************************************************/
/** @brief transform ipix (RING) to ipix (NEST)
 *  @param ipring the ipix (in RING)
 *  @param ipnest the ipix (in NEST) (OUT)
 */
/*****************************************************************************/
void h3cpp_ring2nest(long nside,
                     h3c_ipix_t ipring,
                     h3c_ipix_t *ipnest){
    int order = h3c_order(nside);
#ifdef H3C_INT8
    T_Healpix_Base<int64> healpix(order, NEST);
#else
    Healpix_Base healpix(order, NEST);
#endif

    *ipnest = healpix.ring2nest(ipring);
}

